import numpy as np
import pickle
from read_cifar import read_cifar_batch, split_dataset
import matplotlib.pyplot as plt

def learning_methode(k,dk,learning_rate):
    k2=k-learning_rate*dk
    return(k2)

def learn_once_mse(w1,b1,w2,b2,data,targets,learning_rate):
    a0 = data # the data are the input of the first layer
    z1 = np.matmul(a0, w1) + b1  # input of the hidden layer
    a1 = 1 / (1 + np.exp(-z1))  # output of the hidden layer (sigmoid activation function)
    z2 = np.matmul(a1, w2) + b2  # input of the output layer
    a2 = 1 / (1 + np.exp(-z2))  # output of the output layer (sigmoid activation function)
    predictions = a2  # the predicted values are the outputs of the output layer

    # Compute loss (MSE)
    loss = np.mean(np.square(predictions - targets))

    dc_da2=(2/data.shape[0])*(predictions-targets)
    dc_dz2=dc_da2*(a2*(1-a2))
    dc_dw2=np.dot(np.transpose(a1), dc_dz2)
    dc_db2=np.dot(np.ones((1,dc_dz2.shape[0])),dc_dz2)
    dc_da1=np.dot(dc_dz2,np.transpose(w2))
    dc_dz1=dc_da1*(a1*(1-a1))
    dc_dw1=np.dot(np.transpose(a0), dc_dz1)
    dc_db1=np.dot(np.ones((1,dc_dz1.shape[0])),dc_dz1)

    w1=learning_methode(w1,dc_dw1,learning_rate)
    b1=learning_methode(b1,dc_db1,learning_rate)
    w2=learning_methode(w2,dc_dw2,learning_rate)
    b2=learning_methode(b2,dc_db2,learning_rate)

    # Compute loss (MSE)
    
    return(w1,b1,w2,b2,loss)

def one_hot(label):
    nbr_classe=9
    mat=np.zeros((len(label),nbr_classe))
    for label_indexe,label_im, in enumerate(label):
        mat[label_indexe,label_im-1]=1
    return(mat)

def softmax(y):
    y=np.exp(y)
    v=np.sum(y, axis=1, keepdims=True)
    return(y / v)

def learn_once_cross_entropy(w1,b1,w2,b2,data,labels_train,learning_rate):
    targets = one_hot(labels_train)
    a0 = data
    z1 = np.dot(a0, w1) + b1
    a1 = 1 / (1 + np.exp(-z1))
    z2 = np.dot(a1, w2) + b2
    softa2=softmax(z2)
    predictions=softa2

    loss=np.mean(np.sum(-targets*np.log(predictions),axis=1))
    
    dc_dz2=predictions-targets
    dc_dw2=np.dot(np.transpose(a1), dc_dz2)/data.shape[0]
    dc_db2=np.mean(dc_dz2, axis=0, keepdims=True)

    dc_da1=np.dot(dc_dz2,np.transpose(w2))
    dc_dz1=dc_da1*(1-a1)*a1
    dc_dw1=np.dot(np.transpose(a0), dc_dz1)/data.shape[0]
    dc_db1=np.mean(dc_dz1, axis=0, keepdims=True)

    w1=learning_methode(w1,dc_dw1,learning_rate)
    b1=learning_methode(b1,dc_db1,learning_rate)
    w2=learning_methode(w2,dc_dw2,learning_rate)
    b2=learning_methode(b2,dc_db2,learning_rate)
    
    return(w1,b1,w2,b2,loss)

def accuracy(w1,b1,w2,b2,data,labels):
    a0 = data
    z1 = np.dot(a0, w1) + b1
    a1 = 1 / (1 + np.exp(-z1))
    z2 = np.dot(a1, w2) + b2
    softa2=softmax(z2)
    predictions = softa2
    prediction_2 = np.empty(predictions.shape[0], dtype=int)
    for i, ligne in enumerate(predictions):
        prediction_2[i] = np.argmax(ligne)+1
    indices_egalite = np.where(prediction_2 == labels)[0]
    nombre_indices = len(indices_egalite)
    return(nombre_indices/len(labels))

def train_mlp_cross_entropy(w1,b1,w2,b2,d_train,labels_train,learning_rate,num_epoch):
    train_accuracies,loss_evo=[],[]
    for k in range(num_epoch):
        w1,b1,w2,b2,loss=learn_once_cross_entropy(w1,b1,w2,b2,d_train,labels_train,learning_rate)
        t=accuracy(w1,b1,w2,b2,d_train,labels_train)
        train_accuracies.append(t)
        loss_evo.append(loss)
    return (w1,b1,w2,b2,train_accuracies,loss_evo)

def test_mlp(w1,b1,w2,b2,d_test,labels_test):
    test_accuracy=accuracy(w1,b1,w2,b2,d_test,labels_test)
    return(test_accuracy)

def run_mlp_training(data_train, labels_train, data_test, labels_test,d_h,learning_rate,num_epoch):
    d_in = data_train.shape[1]  # input dimension
    d_out = max(labels_train)  # output dimension (number of neurons of the output layer)

    w1 = np.random.randn(d_in, d_h)  # first layer weights
    b1 = np.zeros((1, d_h))  # first layer biaises
    w2 = np.random.randn(d_h, d_out)  # second layer weights
    b2 = np.zeros((1, d_out))  # second layer biaises

    w1,b1,w2,b2,accuratie_train,loss=train_mlp_cross_entropy(w1,b1,w2,b2,data_train, labels_train,learning_rate,num_epoch)
    test_accuracy=test_mlp(w1,b1,w2,b2,data_test, labels_test)
    return(accuratie_train,loss,test_accuracy)

def unit_test(w1,b1,w2,b2,data_test, labels_test):
    pos=0
    for indexe,image in enumerate(data_test):
        a0 = [image] # the data are the input of the first layer
        z1 = np.matmul(a0, w1) + b1  # input of the hidden layer
        a1 = 1 / (1 + np.exp(-z1))  # output of the hidden layer (sigmoid activation function)
        z2 = np.matmul(a1, w2) + b2  # input of the output layer
        a2 = 1 / (1 + np.exp(-z2))  # output of the output layer (sigmoid activation function)
        predictions = a2  # the predicted values are the outputs of the output layer
        classe = np.argmax(predictions[0])+1
        if classe==labels_test[indexe]:
            pos+=1
    return(pos/len(labels_test))

if __name__ == "__main__":
    d, l = read_cifar_batch("data/cifar-10-batches-py/data_batch_1")
    num_epoch=100
    dh=64
    learning_rate=0.1
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4))
    for k in range(10):
        d_train, l_train, d_test, l_test = split_dataset(d, l, 0.9)
        train_accuracy,loss,test_accuracy=run_mlp_training(d_train, l_train, d_test, l_test,dh,learning_rate,num_epoch)
        ax1.plot(range(num_epoch),loss,label=f"cross-entropy cas {k}")
        ax2.plot(range(num_epoch),train_accuracy,label=f"cross-entropy cas {k}")
        print(k)

    ax1.set_xlabel('epoque')
    ax1.set_ylabel('loss')
    ax1.set_title('evolution de la fonction loss par epoque')
    ax1.legend()

    ax2.set_xlabel('epoque')
    ax2.set_ylabel('accuracy')
    ax2.set_title('evolution de la accuracy')
    ax2.legend()

    plt.tight_layout()
    plt.show()